#!/usr/bin/env node
"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = __importStar(require("fs-extra"));
const minimist_1 = __importDefault(require("minimist"));
const ora_1 = __importDefault(require("ora"));
const path = __importStar(require("path"));
const pretty_ms_1 = __importDefault(require("pretty-ms"));
const chalk_1 = __importDefault(require("chalk"));
const _1 = require(".");
const args = minimist_1.default(process.argv);
const { api, outDir, help } = args;
if (help) {
    console.info(chalk_1.default.cyan('Usage: electron-typescript-definitions --api ../electron-api.json [--out-dir ../electron-out]'));
    process.exit(0);
}
const runner = ora_1.default(chalk_1.default.yellow('Checking argv')).start();
if (typeof api !== 'string') {
    runner.fail(chalk_1.default.red('Missing required --api argument.  "--api ../electron-api.json"'));
    process.exit(1);
}
const resolvedApi = path.isAbsolute(api) ? api : path.resolve(process.cwd(), api);
if (!fs.pathExistsSync(resolvedApi)) {
    runner.fail(`${chalk_1.default.red('Resolved directory does not exist:')} ${chalk_1.default.cyan(resolvedApi)}`);
    process.exit(1);
}
const resolvedOutDir = typeof outDir === 'string'
    ? path.isAbsolute(outDir)
        ? outDir
        : path.resolve(process.cwd(), outDir)
    : process.cwd();
runner.text = chalk_1.default.cyan(`Generating API in directory: ${chalk_1.default.yellow(`"${resolvedOutDir}"`)}`);
const start = Date.now();
const resolvedFilePath = path.resolve(resolvedOutDir, './electron.d.ts');
fs.mkdirp(resolvedOutDir).then(() => _1.generateDefinitions({
    electronApi: require(resolvedApi),
})
    .then(data => fs.writeFile(resolvedFilePath, data))
    .then(() => runner.succeed(`${chalk_1.default.green('Electron Typescript Definitions generated in')} ${chalk_1.default.yellow(`"${resolvedFilePath}"`)} took ${chalk_1.default.cyan(pretty_ms_1.default(Date.now() - start))}`))
    .catch(err => {
    console.error(err);
    process.exit(1);
}));
//# sourceMappingURL=bin.js.map